/*
	KUMAamp project()
	Copyright (C) Mamiya 2000.
	License:LGPL or T98source
	---------------------------------------------------------------------------
	S98 sequencer
	seq_s98.cpp
*/

#include "in_s98.h"

/* ANSI/Windows standard headers */
#include <windows.h>
#include <shlobj.h>
#include <stdio.h>

/* Libraries headers */
#if USE_ZLIB
#include <zlib.h>
#endif

#include <opna.h>

/* Project headers */
#include "../sequencer.h"
#include "in_s98.rh"

#define MAX_CH 16
#define OPNA_CLOCK (3993600)
#define SOUND_55K (55555)
#define SOUND_48K (48000)
#define SOUND_44K (44100)
#define SOUND_24K (24000)
#define SOUND_22K (22050)
#define SOUND_11K (11025)
#define FADE_MAX (64)
#define FADE_SHIFT (20)

/* S98 file header struct */
typedef struct {
	char chk[4];			/* 'S98'(check code) + '1'(format version) */
	int synctime;			/* sync time(ms) */
	int timer_reserve_2;	/* must be zero */
	int compsize;			/* 0:uncompressed 1-:uncompressed size */
	int ofstitle;			/* file offset to title(7bitASCII or SJIS) */
	int ofsdata;			/* file offset to logged data */
	int ofsloop;			/* 0:noloop 1-:file offset to loop point */
} S98HEAD;

/* S98 sequencer struct */
typedef struct {
	SEQUENCER seq;			/* KUMAamp common interface */
	int loopc;				/* zero based loopcounter */
	int loopcount;			/* max loop count */
	int frequency;			/* mixing frequency */
	int fade;				/* 0:normal 1-:fading out(dB/(1 >> FADE_SHIFT)) */
	int fadetime;			/* fadeout time(ms) */
	int voldown;			/* (dB) */

	int volFM;				/* (dB) */
	int volSSG;				/* (dB) */
	int volADPCM;			/* (dB) */
	int volRHYTHM;			/* (dB) */

	S98HEAD head;			/* copy of file header */
	unsigned char *data;	/* data pointer top of file */
	unsigned char *ptr;		/* data pointer current */
	int clock;				/* playtime(ms) */
	int spd1;				/* speedH(samples per 1000usec) */
	int spd2;				/* speedL(samples per    1usec) */
	int left1;				/* stepH (samples per 1000usec) */
	int left2;				/* stepL (samples per    1usec) */
	int left3;				/* stepH fmsync */
	int frequency2;			/* (real) output frequency(Hz) */
	int fmtimersync;		/* 1:OPNA-timer sync mode 0:compatible mode */
	int synctime;			/* synctime(ms) */

	int fadespd;			/* fadeout speed(dB/(1 >> FADE_SHIFT) per 10ms) */
	int isplaying;			/* 1:playing 0:stoped */
} S98SEQ;

/* S98amp setting struct */
static struct {
	/* OPNA rythm wav files path */
	char rhythmpath[MAX_PATH];
	char currentrhythmpath[MAX_PATH];
	int frequency;			/* mixing frequency */
	int priority;			/* play thread priority */
	int loopcount;			/* max loop count */
	int fadetime;			/* fadeout time(ms) */
	int fmtimersync;		/* 1:OPNA-timer sync mode 0:compatible mode */

	int volFM;				/* (dB) */
	int volSSG;				/* (dB) */
	int volADPCM;			/* (dB) */
	int volRHYTHM;			/* (dB) */

	int chenable[MAX_CH];

	HINSTANCE hDllInstance;
	char cfgname[MAX_PATH];
} setting;

typedef struct {
	char *key;
	int num;
} KEYINT;

static FM::OPNA opna;

static KEYINT freqlist[] = {
	{ "55467:FM",	SOUND_55K },
	{ "48000",		SOUND_48K },
	{ "44100",		SOUND_44K },
	{ "24000",		SOUND_24K },
	{ "22050",		SOUND_22K },
	{ "11025",		SOUND_11K },
	{ NULL,			SOUND_55K },
};

static KEYINT priolist[] = {
	{ "Highest",	5 },
	{ "Higher",		4 },
	{ "Normal",		3 },
	{ "Lower",		2 },
	{ "Lowest",		1 },
	{ NULL,			3 },
};

#define PTR__ ((S98SEQ *)(this__->work))
static void Play(SEQUENCER *this__)
{
	PTR__->volFM = PTR__->volSSG = -9999;
	PTR__->volADPCM = PTR__->volRHYTHM = -9999;

	if (lstrcmpi(setting.currentrhythmpath, setting.rhythmpath))
	{
		OutputDebugString(setting.rhythmpath);
		opna.LoadRhythmSample(setting.rhythmpath);
		lstrcpy(setting.currentrhythmpath, setting.rhythmpath);
	}
	PTR__->loopc = 0;
	PTR__->loopcount = setting.loopcount;
	PTR__->fadetime = setting.fadetime;
	PTR__->fade = 0;
	opna.SetRate(OPNA_CLOCK, PTR__->frequency2, (PTR__->frequency > SOUND_44K));

	PTR__->voldown = 0;
	PTR__->fmtimersync = setting.fmtimersync;
	PTR__->clock = 0;
	PTR__->left1 = 0;
	PTR__->left2 = 0;
	PTR__->left3 = 0;
	PTR__->spd1 = (PTR__->frequency2 * PTR__->synctime) / 1000;
	PTR__->spd2 = (PTR__->frequency2 * PTR__->synctime) % 1000;
	PTR__->ptr = PTR__->data + PTR__->head.ofsdata;
	PTR__->isplaying = 1;
	opna.Reset();
	opna.SetReg(0x29, 0x9f);
}
static void Stop(SEQUENCER *this__)
{
	PTR__->loopc = 0;
	PTR__->clock = 0;
	PTR__->left1 = 0;
	PTR__->left2 = 0;
	PTR__->ptr = NULL;
	PTR__->isplaying = 0;
}

#define SETTING_XY(x,y) \
	if (PTR__->y != setting.y) \
	{ \
		PTR__->y = setting.y; \
		opna.x(- PTR__->voldown - PTR__->y); \
	}
#define SETTING_XYA(x,y) \
	if (1) \
	{ \
		PTR__->y = setting.y; \
		opna.x(- PTR__->voldown - PTR__->y); \
	}
#define MIXBUF_SIZE SOUND_48K

static void Mix(SEQUENCER *this__, void *buf, int numsamp)
{
	static int mixbuf[MIXBUF_SIZE*8];
	short *bufp = (short *)buf;
	unsigned char *p = PTR__->ptr;
	if (p == NULL)
	{
		if (bufp != NULL) ZeroMemory(bufp, numsamp*4);
		return;
	}

	if (bufp != NULL)
	{
		SETTING_XY(SetVolumeFM,         volFM);
		SETTING_XY(SetVolumePSG,        volSSG);
		SETTING_XY(SetVolumeADPCM,      volADPCM);
		SETTING_XY(SetVolumeRhythmTotal,volRHYTHM);
	}

	while (bufp == NULL || numsamp)
	{
		int size, *mp;
		while (PTR__->left1 == 0) {
			switch (*p)
			{
				case 0x00:	/* OPNA(m) */
					if (PTR__->fmtimersync && p[1] == 0x27 && (p[2] & 0x30))
					{
						/* sync with fmtimer-reset */
						PTR__->left1 = PTR__->left3;
						PTR__->left3 = 0;
					}
					opna.SetReg(p[1], p[2]);
					p += 3;
					break;
				case 0x01:	/* OPNA(s) */
					opna.SetReg(0x100 | p[1], p[2]);
					p += 3;
					break;
				case 0xff:	/* SYNC(1) */
					if (PTR__->fmtimersync)
					{
						PTR__->left1 = PTR__->left3;
						PTR__->left3 = PTR__->spd1;
					}
					else
					{
						PTR__->left1 = PTR__->spd1;
					}
					PTR__->left2 += PTR__->spd2;
					PTR__->clock += PTR__->synctime;
					p++;
					if (PTR__->fade)
					{
						if (PTR__->fade > (FADE_MAX << FADE_SHIFT))
							PTR__->left1 = -1;
						else
						{
							PTR__->fade += PTR__->fadespd;
							PTR__->voldown = PTR__->fade >> FADE_SHIFT;
							SETTING_XYA(SetVolumeFM,         volFM);
							SETTING_XYA(SetVolumePSG,        volSSG);
							SETTING_XYA(SetVolumeADPCM,      volADPCM);
							SETTING_XYA(SetVolumeRhythmTotal,volRHYTHM);
						}
					}
					break;
				case 0xfe:	/* SYNC(n) */
					{
						int s = 0, n = 0;
						do
						{
							n |= (*(++p) & 0x7f) << s;
							s += 7;
						}
						while (*p & 0x80);
						if (PTR__->fmtimersync)
						{
							PTR__->left1 = PTR__->left3;
							PTR__->left3 = (n + 2) * PTR__->spd1;
						}
						else
						{
							PTR__->left1 = (n + 2) * PTR__->spd1;
						}
						PTR__->left2 += (n + 2) * PTR__->spd2;
						PTR__->clock += (n + 2) * PTR__->synctime;
						p++;
						if (PTR__->fade)
						{
							if (PTR__->fade > (FADE_MAX << FADE_SHIFT))
								PTR__->left1 = -1;
							else
							{
								PTR__->fade += (n + 2) * PTR__->fadespd;
								PTR__->voldown = PTR__->fade >> FADE_SHIFT;
								SETTING_XYA(SetVolumeFM,         volFM);
								SETTING_XYA(SetVolumePSG,        volSSG);
								SETTING_XYA(SetVolumeADPCM,      volADPCM);
								SETTING_XYA(SetVolumeRhythmTotal,volRHYTHM);
#undef SETTING_XY
#undef SETTING_XYA
							}
						}
					}
					break;
				case 0xfd:	/* END */
				default:
					if (PTR__->fmtimersync && PTR__->left3)
					{
						PTR__->left1 = PTR__->left3;
						PTR__->left3 = 0;
						break;
					}
					if (PTR__->head.ofsloop)
					{
						p = PTR__->data + PTR__->head.ofsloop;
						PTR__->loopc++;
					}
					else
						PTR__->left1 = -1;
					break;
			}
		}

		if (PTR__->left1 == -1)
		{
			if (bufp != NULL && numsamp) ZeroMemory(bufp, numsamp*4);
			Stop(this__);
			PTR__->ptr = p;
			return;
		}

		while (PTR__->left2 > 1000)
		{
			PTR__->left2 -= 1000;
			PTR__->left1++;
		}

		if (bufp == NULL)
		{
			PTR__->left1 = 0;
			if (PTR__->clock >= numsamp) break;
		}
		else
		{
			size = min(PTR__->left1, numsamp);
			PTR__->left1   -= size;
			numsamp -= size;
			ZeroMemory(mixbuf, size*8);
			opna.Mix(mixbuf, size);
			size <<= 1;
			mp = mixbuf;
			while (size--)
			{
				int samp;
				samp = *(mp++);
				if (samp > 32767) samp = 32767;
				if (samp < -32768) samp = -32768;
				*(bufp++) = samp;
			}
		}
	}

	PTR__->ptr = p;

	if (PTR__->loopcount == 0 || PTR__->fade != 0) return;
	if (PTR__->loopc < PTR__->loopcount) return;
	if (PTR__->fadetime != 0)
	{
		PTR__->fade = 1;
		PTR__->fadespd = ((FADE_MAX * PTR__->synctime) << FADE_SHIFT) / PTR__->fadetime;
	}
	else
		Stop(this__);
	return;
}

static void Term(SEQUENCER *this__)
{
	if (PTR__->data != NULL) free(PTR__->data);
	free(PTR__);
}
static void SetPosition(SEQUENCER *this__, int time_in_ms)
{

	if (PTR__->clock > time_in_ms) Play(this__);
	Mix(this__, NULL, time_in_ms);
}

static int GetPosition(SEQUENCER *this__)
{
	return PTR__->clock;
}
static int GetRate(SEQUENCER *this__)
{
	PTR__->frequency = setting.frequency;
	PTR__->frequency2 = (PTR__->frequency == SOUND_55K) ? SOUND_44K : PTR__->frequency;
	return PTR__->frequency2;
}
static int GetPriority(SEQUENCER *this__)
{
	return setting.priority;
}

static int GetLoopCount(SEQUENCER *this__)
{
	return PTR__->loopc;
}

static int IsPlaying(SEQUENCER *this__)
{
	return PTR__->isplaying;
}

static int GetLength(SEQUENCER *this__)
{
	int clocks = 0;
	unsigned char *p = PTR__->data + PTR__->head.ofsdata;
	while (1)
	{
		switch (*p)
		{
			case 0x00:	/* OPNA(m) */
			case 0x01:	/* OPNA(s) */
				p += 3;
				break;
			case 0xff:	/* SYNC(1) */
				clocks += 1;
				p++;
				break;
			case 0xfe:	/* SYNC(n) */
				{
					int s = 0, n = 0;
					do
					{
						n |= (*(++p) & 0x7f) << s;
						s += 7;
					}
					while (*p & 0x80);
					clocks += n + 2;
				}
				p++;
				break;
			case 0xfd:	/* END */
			default:
				return clocks * PTR__->synctime;
		}
	}
}

#if SINGLE_BUILD
#define InitS98Sequencer InitSequencer
#define QuitS98Sequencer QuitSequencer
#define AboutS98Sequencer AboutSequencer
#define ConfigS98Sequencer ConfigSequencer
#define getS98FileInfo getFileInfo
#define loadS98File loadFile
#define infoS98Box infoBox
#define isOurS98File isOurFile
#endif

static void WritePrivateProfileInt(LPCSTR sect, LPCSTR key, UINT data, LPCSTR file)
{
	char temp[32], *p;
	temp[30] = '0';
	temp[31] = '\0';
	if (data == 0)
		p = &temp[30];
	else
		for (p = &temp[31]; data != 0; data /= 10)
		{
			*(--p) = '0' + (data % 10);
		}
	WritePrivateProfileString(sect, key, p, file);
}

static void GetSettingString(LPCSTR lpKey, LPCSTR lpDefault, LPSTR lpBuf, DWORD nSize)
{
	GetPrivateProfileString("S98amp", lpKey, lpDefault, lpBuf, nSize, setting.cfgname);
}
static void SetSettingString(LPCSTR lpKey, LPCSTR lpStr)
{
	WritePrivateProfileString("S98amp", lpKey, lpStr, setting.cfgname);
}
static int GetSettingInt(LPCSTR lpKey, int iDefault)
{
	char buf[128], dbuf[128];
	wsprintf(dbuf, "%d", iDefault);
	GetPrivateProfileString("S98amp", lpKey, dbuf, buf, sizeof(buf), setting.cfgname);
	return atoi(buf);
}
static void SetSettingInt(LPCSTR lpKey, int iInt)
{
	char buf[128];
	wsprintf(buf, "%d", iInt);
	WritePrivateProfileString("S98amp", lpKey, buf, setting.cfgname);
}

#include "cfs98amp.h"
#include "cfs98mix.h"

void InitS98Sequencer(HINSTANCE hDllInstance)
{
	char path[MAX_PATH];
	LPTSTR fnoff;

	setting.hDllInstance = hDllInstance;

	GetModuleFileName(hDllInstance, path, MAX_PATH);
	GetFullPathName(path, MAX_PATH, setting.cfgname, &fnoff);
	setting.cfgname[(fnoff - setting.cfgname)] = '\0';
	lstrcat(setting.cfgname, "plugin.ini");

	setting.fmtimersync = 1;

	S98AMPLoadSetting();
	S98MIXLoadSetting();
}

void QuitS98Sequencer(HINSTANCE hDllInstance)
{
}

void AboutS98Sequencer(HWND hwndParent)
{
	MessageBox(hwndParent, "M88 - PC-88 Emulator | FM Sound Generator - Core Unit\n\tCopyright (C) cisc 1998, 2000.\n\nZLIB 1.1.3\n\t(C) 1995-1998 Jean-loup Gailly and Mark Adler\n\nS98amp\n\t(C) 2000 Mamiya\n\t\tmailto:mamiya@proc.org.tohoku.ac.jp\n\t\thttp://www.proc.org.tohoku.ac.jp/befis/", PLUGIN_NAME, MB_OK);
}

void ConfigS98Sequencer(HWND hwndParent)
{
	PROPSHEETPAGE psp[2];
	PROPSHEETHEADER psh;

	ZeroMemory(psp, sizeof(psp));
	psp[0].dwSize = sizeof(PROPSHEETPAGE);
	psp[0].dwFlags = PSP_DEFAULT;
	psp[0].hInstance = setting.hDllInstance;
	psp[0].pszTemplate = MAKEINTRESOURCE(IDD_PSP_S98AMP);
	psp[0].pszIcon = NULL;
	psp[0].pszTitle = NULL;
	psp[0].pfnDlgProc = S98AMPSettingProc;
	psp[0].lParam = 0;
	psp[0].pfnCallback = NULL;
	psp[0].pcRefParent = NULL;

	psp[1].dwSize = sizeof(PROPSHEETPAGE);
	psp[1].dwFlags = PSP_DEFAULT;
	psp[1].hInstance = setting.hDllInstance;
	psp[1].pszTemplate = MAKEINTRESOURCE(IDD_PSP_S98MIX);
	psp[1].pszIcon = NULL;
	psp[1].pszTitle = NULL;
	psp[1].pfnDlgProc = S98MIXSettingProc;
	psp[1].lParam = 0;
	psp[1].pfnCallback = NULL;
	psp[1].pcRefParent = NULL;

	ZeroMemory(&psh, sizeof(psh));
	psh.dwSize = sizeof(PROPSHEETHEADER);
	psh.dwFlags = PSH_NOAPPLYNOW | PSH_PROPSHEETPAGE;
	psh.hwndParent = hwndParent;
	psh.hInstance = setting.hDllInstance;
	psh.pszIcon = NULL;
	psh.pszCaption = (LPSTR) "S98amp setting";
	psh.nPages = sizeof(psp) / sizeof(PROPSHEETPAGE);
	psh.nStartPage = 0;
	psh.ppsp = (LPCPROPSHEETPAGE) &psp;
	psh.pfnCallback = NULL;
	PropertySheet(&psh);
}

SEQUENCER *loadS98File(char *fn)
{
	FILE *fp = NULL;
	S98SEQ *p = NULL;
#if USE_ZLIB
	gzFile gzfp = NULL;
#endif
	while (1) {
		size_t size;
		p = (S98SEQ *) malloc(sizeof(S98SEQ));
		if (p == NULL) break;;
		p->seq.work = p;
		p->data = NULL;
		fp = fopen(fn, "rb");
		if (fp == NULL) break;
		fseek(fp, 0, SEEK_END);
		size = ftell(fp);
		fseek(fp, 0, SEEK_SET);
		p->data = (unsigned char *)malloc(size);
		if (p->data == NULL) break;
		if (size != fread(p->data, 1, size, fp)) break;
		fclose(fp); fp = NULL;
#if USE_ZLIB
		if (p->data[0] == 0x1f && p->data[1] == 0x8b) {
			int dessize = 0;
			unsigned char *des;
			gzfp = gzopen(fn, "rb");
			if (gzfp == NULL) break;
			while (!gzeof(gzfp)) {
				size += 16 * 1024;
				des = (unsigned char *)realloc(p->data, size);
				if (des == NULL) break;
				p->data = des;
				dessize = gztell(gzfp);
				if (gzread(gzfp, des + dessize, size - dessize) == -1) break;
				dessize = gztell(gzfp);
			}
			if (dessize < 16) break;
			if (des[0] != 'S' || des[1] != '9' || des[2] != '8') {
				if (dessize < 512 || des[512+0] != 'S' || des[512+1] != '9' || des[512+2] != '8') {
					break;
				}
				/* TAR.GZ ??? */
				dessize -= 512;
				MoveMemory(des, des + 512, dessize);
			}
			des = (unsigned char *)realloc(p->data, dessize);
			if (des != NULL) p->data = des;
			gzclose(gzfp); gzfp = NULL;
		}
#endif

		CopyMemory(&p->head, p->data, sizeof(S98HEAD));

		if (p->head.chk[0] != 'S' || p->head.chk[1] != '9' || p->head.chk[2] != '8') break;
		/* version */
		if (p->head.chk[3] < '0' || '1' < p->head.chk[3]) break;

		if (p->head.compsize) {
#if USE_ZLIB
			uLongf dessize;
			unsigned char *des;
			des = (unsigned char *)malloc(p->head.compsize + p->head.ofsdata);
			if (des == NULL) break;
			CopyMemory(des, p->data, p->head.ofsdata);
			if (Z_OK != uncompress(des + p->head.ofsdata, &dessize, p->data + p->head.ofsdata, size - p->head.ofsdata)) {
				free(des);
				break;
			}
			free(p->data);
			p->data = des;
#else
			break;	/* NOT SUPPORT */
#endif
		}
		p->synctime = p->head.synctime ? p->head.synctime : 10;
		p->clock = 0;
		p->left1 = 0;
		p->ptr = NULL;
		p->loopc = 0;
		p->isplaying = 0;

		p->seq.Play = Play;
		p->seq.Stop = Stop;
		p->seq.Mix = Mix;
		p->seq.Term = Term;
		p->seq.SetPosition = SetPosition;
		p->seq.GetPosition = GetPosition;
		p->seq.GetRate = GetRate;
		p->seq.GetPriority = GetPriority;
		p->seq.GetLoopCount = GetLoopCount;
		p->seq.IsPlaying = IsPlaying;
		return &p->seq;
	}
#if USE_ZLIB
	if (gzfp != NULL) gzclose(gzfp);
#endif
	if (fp != NULL) fclose(fp);
	if (p != NULL) Term(&p->seq);
	return NULL;
}

int getS98FileInfo(char *fn, char *title, int *length_in_ms)
{
	SEQUENCER *this__;
	this__ = loadS98File(fn);
	if (this__ == NULL) return FALSE;
	if (title != NULL)
	{
		if (PTR__->head.ofstitle)
			lstrcpy(title, (char *)PTR__->data + PTR__->head.ofstitle);
		else
		{
			char *p, *p2;
			for (p2 = p = fn;1;)
			{
				char *prev;
				if (*p == '/' || *p == '\\' || *p == ':') p2 = p + 1;
				p = CharNext(prev = p);
				if (p == prev) break;
			}
			lstrcpy(title, "[ ");
			lstrcat(title, p2);
			lstrcat(title, " ]");
		}
	}
	if (length_in_ms != NULL)
		*length_in_ms = GetLength(this__);
	Term(this__);
	return true;
}

int infoS98Box(char *fn, HWND hwndParent)
{
	return 0;
}

int isOurS98File(char *fn)
{
	if (fn)
	{
		int l = lstrlen(fn) - 7;
		if (l >= 0 && lstrcmpi(fn + l, ".S98.GZ") == 0) return 1;
	}
	return 0;
}
